# Antivir scanners

from basic import *
from dazuko import *
from clamav import *
from avg import avgd
from bdc import bdc
from nod import *
from sophie import sophie
from kav import *
from icap import *
from sanitize import *



'''
Advanced logger for sagator, version 0.2
 
(c) 2003-2004 Jan ONDREJ (SAL) <ondrejj(at)salstar.sk>

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
                                                                                
'''
                                                                                
from avlib import *
from match import match_any
from report import report
                                                                                
__all__=['log']

class log(report):
  '''
  Advanced logger interscanner.
  
  usage: log(logto,format,scanners...)
  
  where: logto is a string, which defines filename to store these data
           If logto is an integer, it defines log level in standard
           log file.
         format is a string, which defines data format

  This scanner can be used to log some special data.
  You can use these variables in format string:
    $LEVEL		- detected virus level
    $VIRNAME		- name of detected virus, empty if CLEAN
    $STATUS		- if mail is dropped, rejected, ...
    $QNAME		- quarantine file name
    $SCANNER_NAME	- scanner which reported this virus
    $SENDER		- scanned message sender
    $RECIPIENTS		- scanned message recipients
    $SIZE               - email size
  
  You also can use log.FORMAT for default format.
  '''
  name='log()'
  # default format
  FORMAT="level='$LEVEL', virname='$VIRNAME', size='$SIZE', sender='$SENDER', recipients='$RECIPIENTS'\n"
  def __init__(self,logto,format,*scanners):
      self.logto=logto
      self.format=format
      match_any.__init__(self,scanners)
  def repl_vars(self,level,detected):
      return self.replace_tmpl(self.format,{
        'LEVEL':	str(level),
        'VIRNAME':	detected,
        'STATUS':	globals.ACTION,
        'QNAME':	globals.QFNAME,
        'SCANNER_NAME':	globals.found_by.name,
        'SENDER':	mail.sender.strip(),
        'RECIPIENTS':	string.join(mail.recip,','),
        'SIZE':		str(len(mail.data))
      })
  def scanstream(self,stream):
      level,detected,virlist=match_any.scanstream(self,stream)
      if type(self.logto)==type(1):
        debug.echo(self.logto,self.repl_vars(level,detected).split())
      else:
        open(self.logto,'a').write(self.repl_vars(level,detected))
      return level,detected,virlist



#!/usr/bin/python2.3

'''Sagator Back from Quarantine script.
 
This program can be used to requeue an quarantined mail from
quarantine to queue/user mailbox.
 
(c) 2003-2004 Jan ONDREJ (SAL) <ondrejj(at)salstar.sk>

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

Usage:
  sqback [--to RCPT] quarantined_file1 [quarantined_files...]
'''

import sys,socket,re
from default import SMTP_SERVER
from config import SMTP_SERVER

def showhelp():
    print "sqback usage:"
    print "  sqback [--to RCPT] quarantined_file1 [quarantined_files...]"
    sys.exit()

rcptto=''
args=[]
n=1
while n<len(sys.argv):
  if sys.argv[n]=='--to':
    rcptto=sys.argv[n+1]
    n+=2
    continue
  if sys.argv[n][:2]=='--':
    print "Unknown parameter:",sys.argv[n]
    showhelp()
  else:
    args.append(sys.argv[n])
    n+=1

if len(args)==0:
  showhelp()

try:
  for fname in args:
    conn=socket.socket(socket.AF_INET,socket.SOCK_STREAM)
    conn.connect(SMTP_SERVER)
    print "Connection to: "+str(SMTP_SERVER)+" established"
    f=open(fname,'r')
    r_rcptto=re.compile("^RCPT TO:.*[\r]*$",re.IGNORECASE|re.MULTILINE)
    r_data=re.compile("^DATA[\r]*$",re.IGNORECASE|re.MULTILINE)
    if rcptto:
      buf=f.read()
      buf=r_rcptto.sub("RCPT TO: "+rcptto,buf,1)
      while 1:
        e=r_rcptto.search(buf).end()
        s=r_data.search(buf).start()
        while buf[e] in ['\r','\n']:
          e+=1
        if e<s:
          buf=buf[:e]+buf[s:]
        else:
          break
      conn.sendall(buf)
    else:
      conn.sendall(f.read())
    f.close()
    while 1:
      try:
        r=conn.recv(10240)
        if r=='': break
        print r.strip()
      except:
        print "Connection closed."
        break
except IOError,(ec,es):
  print "IOError:", es+":" , fname



'''
clamav module, version 0.6.0

(c) 2003 Jan ONDREJ (SAL) <ondrejj(at)salstar.sk>
                                                                                
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

'''

from avlib import *
import socket,re,string

__all__=['clamscan','clamd','libclam']

def f_clamscandir(fda,scanner):
    level,rc,ret=0.0,0,[]
    for fdir in fda:
      pf=popen(scanner+[safe.fn(fdir)])
      pf.tocmd.close()
      rc=pf.wait()
      for line in pf.readlines():
        if re.search(" FOUND",line):
          ret.append(line)
          level+=1.0
      pf.fromcmd.close()
      debug.echo(4,"CLAMSCANF: RET: ",rc,ret)
    if rc==0:
      level,vir=0.0,''
    elif rc==1:
      l=re.search(": (.*) FOUND",string.rstrip(ret[0]))
      if l:
        vir=l.group(1)
      else:
        debug.echo(0,'CLAMSCANF: ERROR: ',ret)
        raise ScannerError,'clamscandir: '+str(ret)
    else:
      raise ScannerError,'clamscan: returned error code: '+str(rc)+str(ret)
    return level,vir,ret

def f_clamd(stream,scanner):
    if type(scanner)==type('string'): # local socket
      s=socket.socket(socket.AF_UNIX,socket.SOCK_STREAM)
      socket_settimeout(s,120)
      s.connect(scanner)
    else:
      s=socket.socket(socket.AF_INET,socket.SOCK_STREAM)
      socket_settimeout(s,120)
      s.connect((scanner[0],scanner[1]))
    f=s.makefile('rw',BUFSIZE)
    s.sendall("STREAM\n")
    rl=f.readline()
    l=re.search('^PORT ([0-9]*)',rl)
    if l:
      # Send stream
      ss=socket.socket(socket.AF_INET,socket.SOCK_STREAM)
      socket_settimeout(ss,120)
      if type(scanner)==type('string'): # local socket
        ss.connect(('127.0.0.1',string.atoi(l.group(1))))
      else:
        ss.connect((scanner[0],string.atoi(l.group(1))))
      ss.sendall(fromhdr())
      ss.sendall(stream)
      ss.shutdown(2)
      ss.close()
      # Read reply
      level=0.0
      ret=[]
      for line in f.readlines():
        if not re.search(": OK$",string.rstrip(line)):
          ret.append(line)
      f.close()
      s.shutdown(2)
      s.close()
      if ret==[]:
        level,vir=0.0,''
      else:
        l=re.search(": (.*) FOUND$",string.rstrip(ret[0]))
        if l:
          vir=l.group(1)
          level=1.0
        else:
          debug.echo(0,'CLAMD: ERROR: ',ret)
          raise ScannerError,'clamd: '+str(ret)
      return level,vir,ret
    else:
      raise ScannerError,'Protocol error, status: '+rl

class clamscan(ascanner):
  '''
  clamscan() realscanner - clamav scanner
  
  Usage: clamscan(['/usr/bin/clamscan','--stdout','--infected',
                   '--disable-summary','-r','--mbox'])

  Where: in [] are command line parameters for clamscan binary

  This scanners is a realscanner, which can be used to scan
  for viruses. ClamAV is a free program under GPL and may be freely
  used. Its database is very well.
  
  It can be easily used in chroot. mkchroot.sh from sagator package
  automatically integrates it into chroot environment.

  WARNING: If you are using clamscan from chroot, you need to configure
           freshclam to update chrooted copy of database!!!
  '''
  name='clamscan()'
  def __init__(self,arg):
      self.arg=arg
  def scanfile(self,files,dirname=''):
      level,detected,virlist=f_clamscandir([dirname],self.arg)
      return level,detected,virlist

class clamd(ascanner):
  '''
  clamd realscanner
  
  Usage: clamd(['localhost',3310])
     or: clamd('/var/spool/vscan/clamd')

  Where: 1st definition defines a clam daemon on host localhost,
             port 3310 (it is the default)
         2nd definition defines access to clamd via UNIX socket

  This scanners is a realscanner, which can be used to scan
  for viruses. ClamAV is a free program under GPL and may be freely
  used. Its database is very well.

  If you need a chrooted clamd, copy it into chroot and start it
  from here. If you are not familiar with it, use clamscan scanner
  from this module. It easily does chroot support, but it
  is not powerfull.
  '''
  name='clamd()'
  def __init__(self,arg):
      self.arg=arg
  def scanstream(self,stream):
      level,detected,virlist=f_clamd(stream,self.arg)
      return level,detected,virlist

try:
  import libclamav

  class libclam(ascanner):
    '''
    libclam realscanner - uses libclamavmodule python library
    
    Usage: libclam([options[,limits]])
    
    Where: options is an number, whic defines libclam options,
             see clamav documentation for more info
           limits is an tuple, which defines limits for clamav
             decompressor, see clamav doc. for more info
    
    This scanners is a realscanner, which can be used to scan
    for viruses. ClamAV is a free program under GPL and may be freely
    used. Its database is very well.
    
    WARNING: Streamscanner ignores options and limits!

    WARNING: Streamscanner can't scan mailboxes and archives.
             Use parsemail() for emails anb if you need also
             archive support, use decompress() interscanner.

    WARNING: If you are using clam from chroot, you need to configure
             freshclam to update chrooted copy of database!!!
    '''
    name='libclam()'
    CL_RAW=libclamav.CL_RAW
    CL_ARCHIVE=libclamav.CL_ARCHIVE
    CL_MAIL=libclamav.CL_MAIL
    CL_DISABLERAR=libclamav.CL_DISABLERAR
    CL_OLE2=libclamav.CL_OLE2
    CL_ENCRYPTED=libclamav.CL_ENCRYPTED
    CL_DEFAULT=libclamav.CL_DEFAULT
    def __init__(self,options=libclamav.CL_DEFAULT,limits={}):
        self.av=None # do not load database before chroot!
        self.options=options
        self.limits=limits
    def reinit(self):
        if not self.av:
          self.av=libclamav.clamav()
          self.av.setlimits(self.limits)
        else:
          if self.av.reload()==0:
            return
        debug.echo(2,self.name,": Loaded virpatterns: ",self.av.virnum)
    def scanstream(self,stream):
        try:
          v=self.av.scanbuff(stream)
          if v:
            return 1.0,v,[v]
        except RuntimeError,es:
          debug.echo(1,"libclam(): "+es)
          raise
        return 0.0,'',[]
    def scanfile(self,files,dir=''):
        for fname in files:
          fd=safe.osopen(fname,os.O_RDONLY)
          try:
            v=self.av.scandesc(fd,self.options)
            os.close(fd)
            if v:
              mail.addheader('X-Sagator-LibClamAV',v)
              return 1.0,v,[fname+": "+v]
          except:
            os.close(fd)
            raise
        return 0.0,'',[]
except ImportError,es:
  class libclam(ascanner):
    '''
    This method is undefined, because libclamavmodule is not found.
    Please recompile sagator, if you really need it.
    '''
    name='libclam()'
    errorstring="Importing of libclamavmodule failed!"
    def __init__(self):
        pass
    def scanstream(self,stream):
        debug.echo(1,self.name+": scanstream(): "+self.errorstring," [",es,"]")
        raise ScannerError,self.errorstring
    def scanfile(self,files,dirname=''):
        debug.echo(1,self.name+": scanstream(): "+self.errorstring," [",es,"]")
        raise ScannerError,self.errorstring



'''
Dazuko scanner for sagator, version 0.1

(c) 2003-2004 Jan ONDREJ (SAL) <ondrejj(at)salstar.sk>

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
    
'''

from avlib import *
import re,string,os

__all__=['dazuko']

class dazuko(ascanner):
  '''
  Scanner which uses dazuko directory to scan for viruses.
  
  Usage: dazuko('/dazuko_prefix','logfile')

  Where: '/dazuko_prefix' is a string, which defines directory
            and a part of filename to store scanned files,
            by default '/tmp/dazuko'
         'logfile' is a string, which defines full path and filename
            of logfile, by default not logfile used
         'regexp' is a string, which defines how to parse logfile
  '''
  name='dazuko()'
  def __init__(self,dazuko_prefix,logfile='',regexp=''):
      self.prefix=dazuko_prefix
      self.logfile=logfile
      self.reg=re.compile(regexp,re.IGNORECASE)
  def scanstream(self,stream):
      level,detected,virlist=0.0,'',[]
      # open logfile
      if self.logfile:
        log=open(self.logfile,'r')
        log.seek(0,2) # end of file
      # save email
      t=mktemp(self.prefix,'.eml','w',0600).autorm()
      try:
        t.f.write(fromhdr())
        t.f.write(stream)
        t.f.flush()
        open(t.root_name,'r') # IOError1 raised when virus found
      except IOError,(ec,es):
        if ec==1:
          detected='UNKNOWN'
          level+=1.0
          if log:
            l=1
            while l:
              l=log.readline()
              reg1=self.reg.search(l)
              if reg1:
                if reg1.group(1) in [t.root_name,t.name]:
                  detected=reg1.group(2)
                  virlist.append(l)
                  debug.echo(6,"NOD2d: ",l)
        else:
          raise
      # close log
      log.close()
      return level,detected,virlist



'''
Basic interscanners for sagator, version 0.6.0

(c) 2003-2004 Jan ONDREJ (SAL) <ondrejj(at)salstar.sk>

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
    
'''

from avlib import *
import os

__all__=['file2stream','stream2file','f2s','s2f','stream2mbox']

#####################################################################
### convertors

class file2stream(interscanner):
  '''
  file2stream interscanner
  
  Usage: file2stream(scanner())
  
  Where: scanner() is an interscanner or realscanner.
           It must be a streamscanner.
  
  This scanner converts a filescanner input into streamscanner.
  '''
  name='file2stream()'
  def __init__(self,scanner):
      self.rename([scanner])
      self.scanner=scanner
  def scanfile(self,files,dir=''):
      for fn in files:
        f=safe.open(fn)
        self.scanner.prescan()
        level,vir,ret=self.scanner.scanstream(f.read())
        self.scanner.postscan(level,vir,ret)
        f.close()
        if vir:
          return level,vir,ret
      return level,'',[]

class f2s(file2stream):
  '''An alias for file2stream interscanner'''
  name='f2s()'

class stream2file(interscanner):
  '''
  stream2file interscanner
  
  Usage: stream2file(scanner())
  
  Where: scanner() is a interscanner or realscanner.
           It must be a filescanner.
  
  You can use it to save streams into files.
  '''
  name='stream2file()'
  is_multistream=1
  def __init__(self,scanner,addheader=0):
      self.rename([scanner])
      self.scanner=scanner
      self.addheader=addheader
      self.filename=''
      self.mboxdir=''
      self.filenames=[]
      self.counter=1
  def destroy(self):
      if self.mboxdir:
        if debug.debug_level<10:
          debug.echo(5,"s2f: destroy():",self.mboxdir)
          rdel([self.mboxdir])
        else:
          debug.echo(4,"s2f: unfinished destroy:",self.mboxdir)
      self.scanner.destroy()
      self.filenames=[]
      self.mboxdir=''
  def creatembox(self,f,data,addheader=0):
      try:
        if addheader:
          f.write(fromhdr())
        f.write(data)
        f.close()
      except os.error,(ec,es):
        debug.echo(0,'creatembox: ERROR: Open: ',es)
        raise
      except IOError,(ec,es):
        debug.echo(0,'creatembox: ERROR: IO: ',es)
        raise
  def addstream(self,stream):
      if stream:
        if self.mboxdir=='':
          self.mbox=mktemp('/tmp/s2f-','.mbd','d',0700)
          self.mboxdir=self.mbox.name
        if self.filename=='':
          self.filename='part-%06d' % (self.counter)
          self.counter=self.counter+1
        mboxfile=mktemp(self.mboxdir+'/','-'+self.filename,'w',0600)
        self.creatembox(mboxfile.f,stream,self.addheader)
        self.filenames.append(mboxfile.name)
  def scanstream(self,stream):
      level,vir,ret=0.0,'',[]
      self.addstream(stream)
      try:
        self.scanner.prescan()
        level,vir,ret=self.scanner.scanfile(self.filenames,self.mboxdir)
        self.scanner.postscan(level,vir,ret)
      except:
        raise
      return level,vir,ret
  def scanmultistream(self,streams):
      level,vir,ret=0.0,'',[]
      if type(streams)==type({}):
        for filename,stream in streams.items():
          self.filename=filename
          self.addstream(stream)
      else:
        for stream in streams:
          self.addstream(stream)
      try:
        self.scanner.prescan()
        level,vir,ret=self.scanner.scanfile(self.filenames,self.mboxdir)
        self.scanner.postscan(level,vir,ret)
      except:
        raise
      return level,vir,ret

class s2f(stream2file):
  '''An alias for stream2file interscanner.'''
  name='s2f()'

class stream2mbox(stream2file):
  '''An alias for stream2file(...,1) interscanner.'''
  name='stream2mbox()'
  def __init__(self,scanner):
      stream2file.__init__(self,scanner,1)



'''
bdc module - Bitdefender support, version 0.1

(c) 2003-2004 Jan ONDREJ (SAL) <ondrejj(at)salstar.sk>
                                                                                
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

'''

from avlib import *
import os,re,string

__all__=['bdc']

class bdc(ascanner):
  '''
  Bitdefender realscanner
  
  Usage: bdc(command)

  Where: command is a array which defines command with parameters.
           By default: ['/opt/bdc/bdc','--mail','--all','--arc']

  This scanners is a realscanner, which can be used to scan
  for viruses. It parses Bitdefender output.

  Warning: If you are using bdc in chroot, don't forget to mount
           /proc directory under chroot.
  '''
  name='bdc()'
  def __init__(self,command=['/opt/bdc/bdc','--mail','--all','--arc']):
      self.command=command
  def scanfile(self,files,dir=''):
      level,detected,virlist,vir_count,ret=0.0,'',[],0,[]
      if dir:
        pf=popen(self.command+[safe.fn(dir)])
      else:
        safe_files=[]
        for i in files:
          safe_files.append(safe.fn(i))
        print safe_files
        pf=popen(self.command+safe_files)
      pf.tocmd.close()
      rc=pf.wait()
      # read lines from command
      while 1:
        line=pf.readline()
        if line=='': break
        if re.search("^Results:",line):
          while 1:
            line=pf.readline()
            if line=='': break
            if re.search("^Identified viruses:",line):
              vir_count=int(line[19:].strip())
          break
        if re.search("^Error: ",line):
          debug.echo(3,"ERROR: bdc(): ",line.strip())
        re1=re.search(" (inf|susp)ected: (.*)$",line.rstrip())
        if re1:
          if re1.group(1)=="inf": # if it is an indentified virus
            detected=re1.group(2)
            level+=1.0
          else:
            level+=0.9 # only suspected
            if detected=='': # if it is suspected, but no ident. virus
              detected=re1.group(2)
          ret.append(line)
      pf.fromcmd.close()
      debug.echo(4,"BDC: RET: ",[[vir_count,rc]],ret)
      return level,detected,virlist



'''
spamassassin module, version 0.6.1

(c) 2003 Jan ONDREJ (SAL) <ondrejj(at)salstar.sk>
                                                                                
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

'''

from avlib import *
import re,socket

__all__=['spamassassin','spamassassind']

class spamassassin(ascanner):
  '''
  SpamAssassin scanner
  
  Usage: spamassassin(['/usr/bin/spamassassin','-e'],SPAMSTRING,SA_MAX_FILE)
  
  Where: in [] is a path and argumets of spamassassin binary
         SPAMSTRING is a string, which define a spamassassin report,
           which is checked. If it is found, mail is assigned as 'SPAM'
           (by default all reports with exit status!=0 are reported as spam).
         SA_MAX_FILE is a number, which defines maximum file size,
           which can be tested. If a mail (with header) is larger,
           it is not checked (by default 500000).

  This scanner scans for spams. If a spam is returned, a string
  SPAM is returned as virus name. This scanner uses a spamassassin
  binary. For better performance, configure spamd and use
  spamassassind.
  
  WARNING! You need to install spamassassin into chroot manually.
  mkchroot.sh script does it not. If you want to use spamassasin
  with chroot and don't want to make a copy in chroot, use
  a spamassassind scanner, but remember, that scanning for spams
  are not done in chroot environment.
  
  WARNING! This scanner is not fully compatible with header adding.
  Some data from headers are not added!
  '''
  is_spamscan=1
  name='SpamAssassin()'
  def __init__(self,arg,spamstring='',sa_max_file=500000):
      self.arg=arg
      self.SPAMSTRING=spamstring
      self.SA_MAX_FILE=sa_max_file
  def scanstream(self,stream):
      exitstatus=''
      if len(stream)>self.SA_MAX_FILE:
        return 0.0,'',['File too long to test it. Size: '+str(len(stream))]
      pf=popen(self.arg)
      pf.write(stream)
      pf.tocmd.close()
      pf.wait()
      level,ret,exitstatus,lastspamline=0.0,[],'',0
      for line in pf.readlines():
        if re.search("^$",line):
          break
        if re.search("^\t",line):
          if lastspamline:
            ret.append(ret.pop(-1)+line)
        else:
          if re.search("^X-Spam-",line):
            ret.append(line)
            lastspamline=1
            if self.SPAMSTRING:
              if re.search(self.SPAMSTRING,line):
                exitstatus='SPAM'
            reg2=re.search("X-Spam-Status: Yes, hits=([^, ]*)[, ]",line)
            if reg2:
              level=float(reg2.group(1))
          else:
            lastspamline=0
      for i in ret:
        mail.addheader(i)
      pf.fromcmd.close()
      if (pf.exitstatus>0) and (self.SPAMSTRING==''):
        exitstatus='SPAM'
      return level,exitstatus,ret

class spamassassind(ascanner):
  '''
  SpamAssassinD scanner
  
  Usage: spamassassind(['localhost',783],REQ_SPAM_LEVEL,SA_MAX_FILE)

  Where: ['localhost',783] are the host and port of spamd.
         if REQ_SPAM_LEVEL is less than spam hits, virus is returned,
           if it is -1 (default), spam level from spamassassin config is used
         for SA_MAX_FILE see spamassassin scanner description.
  
  This scanner scans for spams. If a spam is returned, a string
  SPAM is returned as virus name.
  This scanner uses a spamassassind daemon. If you are not familiar with it,
  you also can use spamassassin binary. Look at spamassassin scanner.
  '''
  is_spamscan=1
  name='SpamAssassinD()'
  reg_x_spam=re.compile('^(X-Spam-.*?)\r?\n[^ \t]',re.M|re.S).search
  reg_eoh=re.compile('^(.*?\r?\n\r?\n)',re.M|re.S).search
  reg_status=re.compile(
    '^X-Spam-Status: (Yes|No), (score|hits)=([-0-9.]*) (required)=([-0-9.]*) ').search
  reg_hdr_body=re.compile('^([^:]*):[ \t]*(.*)',re.M|re.S).search
  def __init__(self,arg,reqspamlevel=-1,sa_max_file=500000):
      self.arg=arg
      self.REQSPAMLEVEL=reqspamlevel
      self.SA_MAX_FILE=sa_max_file
  def scanstream(self,stream):
      if len(stream)>self.SA_MAX_FILE:
        return 0.0,'',['File too long to test it. Size: '+str(len(stream))]
      s=socket.socket(socket.AF_INET,socket.SOCK_STREAM)
      socket_settimeout(s,120)
      s.connect((self.arg[0],self.arg[1]))
      f=s.makefile('rw',BUFSIZE)
      s.sendall("PROCESS SPAMC/1.2\r\n")
      s.sendall("Content-length: "+str(len(stream))+"\r\n")
      s.sendall("User: vscan\r\n\r\n")
      s.sendall(stream)
      s.shutdown(1)
      s.close()
      output1=f.readline().rstrip()
      debug.echo(4,"Spamd output: ",[[len(stream)]],str(output1))
      # spamd returned any error?
      if not re.search('^SPAMD/[0-9.]* 0 ',output1):
        raise ScannerError,'SpamAssassind: '+output1
      # workaround for 8bit mails
      if re.search('^SPAMD/[0-9.]* 76 ',output1):
        debug.echo(3,"ERROR: spamd(): Spamd returned: "+output1)
        return 0.0,'',[]
      # spamd returned EX_OK
      content_length=0
      while True:
        output2=f.readline()
        if not output2.strip(): # end of spamd header?
          break
        reg1=re.search('^Content-length: ([0-9]*)$',string.rstrip(output2))
        if reg1:
          content_length=int(reg1.group(1))
        else:
          debug.echo(0,"ERROR: spamd(): unknown header: %s" % output2)
      # read reported message and close connection
      output=f.read(content_length)
      f.close()
      # process returned email's header
      try:
        header=self.reg_eoh(output).group(1)
      except:
        header=output
      pos=0
      ret=[]
      spam_yn,spam_score,spam_req='No',0.0,1000.0
      while pos<len(header):
        reg2=self.reg_x_spam(header[pos:])
        if reg2:
          line=reg2.group(1).rstrip()
          pos+=reg2.end(1)
          ret.append(line)
          reg3=self.reg_status(line)
          if reg3:
            spam_yn=reg3.group(1)
            spam_score=float(reg3.group(3))
            spam_req=float(reg3.group(5))
          if line[:10]=='X-Spam-Rep':
            debug.echo(5,"spamd(): ",line)
          else:
            debug.echo(4,"spamd(): ",line)
          #reg4=self.reg_hdr_body(line)
          mail.xheader+=line+'\r\n' # not safe, use mail.addheader instead!
        else:
          break
      if self.REQSPAMLEVEL>=0:
        req_level=self.REQSPAMLEVEL
      else:
        req_level=spam_req
      if spam_score>=req_level:
        detected="SPAM"
      else:
        detected=''
      debug.echo(3,"Spamd status: [%s], score=%f/%f" \
        % (detected,spam_score,spam_req))
      return float(spam_score/req_level),detected,ret



#!/usr/bin/python2.3

'''sgscan - sagator's mailbox/email scanner

(c) 2003-2004 Jan ONDREJ (SAL) <ondrejj(at)salstar.sk>

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
'''

SAVECLEAN=''
AV_ONLY=0
VERBOSE=0

import sys,os,fcntl,re,time,struct,string
from cStringIO import StringIO
from aglib import *

# load config
from default import *
try:
  from config import *
except ImportError,es:
  if str(es)[-6:]=="config":
    print "ERROR! Config file not found! Exiting now."
    sys.exit(1)
  else:
    raise

def mailhdr(str):
    return str[0:5]=="From "

def scan(data,fn,mailcount=-1):
    mail.data=data
    mail.xheader=''
    debug.echo(2, "Object: ",fn)
    scanarr,scannames=[],'None'
    globals.reset()
    if mailhdr(fn):
      fn2=fn.split()
      fn2=string.join([fn2[1]]+fn2[3:])
    else:
      fn2=fn
    level=0.0
    for scnr in SCANNERS:
      if (AV_ONLY>0) and (scnr.get('is_spamscan')>0): # if it contain a spamscan
        debug.echo(4,"Spam scanner skipped: ",scnr.name)
      else:
        l,detected,virlist,scan_reply,err=do_scan(scnr,os.path.basename(fn))
        if err:
          if mailcount>0:
            print "%2d: %s: %s" % (mailcount,fn2,err)
          else:
            print "%s: %s" % (fn,err)
          return False
        level+=l
        if scan_reply:
          debug.echo(2, scan_reply)
        else:
          scanarr=scanarr+[scnr.name]
          scannames=string.join(scanarr,' ')
        if is_infected(l,detected):
          if mailcount>0:
            print "%2d: %s: %s [%s,%0.2f]" % (mailcount,fn2,detected,
              globals.found_by.name,level)
          else:
            print "%s: %s [%s,%0.2f]" % (fn,detected,
              globals.found_by.name,level)
          return True
    if VERBOSE:
      if mailcount>0:
        print "%2d: %s: CLEAN" % (mailcount,fn2)
      else:
        print fn+": CLEAN"
    return False

debug.set_logfile('-')
debug.set_level(0)
globals.scan_only=True
# parse command line arguments
files=[]
n=1
while n<len(sys.argv):
  if sys.argv[n]=="--help":
    print "sgscan - sagator's file scanner"
    print ""
    print "Usage: sgscan [--config F] [--debug X] \\"
    print "              [--clean] [--verbose] [--av-only] files..."
    print ""
    print "Params: --help        this help"
    print "        --config f    load \"f\" as alternate config (without .py extension)"
    print "        --logfile l   filename for logging ('-' for stdout)"
    print "        --debug l     set debug level to l"
    print "        --av-only     don't run antispam test, only antivir tests"
    print "        --clean       clean infected emails from mailboxes"
    print "        --verbose     show also CLEAN emails"
    sys.exit(0)
  elif sys.argv[n]=='--config':
    n+=1
    try:
      exec("from %s import *" % sys.argv[n])
    except ImportError,es:
      print "ImportError:",es
  elif sys.argv[n]=="--clean":
    SAVECLEAN=".clean"
  elif sys.argv[n]=="--av-only":
    AV_ONLY=1
  elif sys.argv[n]=="--verbose":
    VERBOSE=1
  elif sys.argv[n]=="--debug":
    n+=1
    debug.set_level(int(sys.argv[n]))
  else:
    files.append(sys.argv[n])
  n+=1

# reinit scanners
for scnr in SCANNERS:
  scnr.reinit()
safe.ROOT_PATH=CHROOT

cwd=os.getcwd()
total,infected=0,0
time0=time.mktime(time.localtime())
for fn in files:
  while 1:
    try:
      try:
        os.chdir(cwd)
        f=open(fn,"r+")
      except IOError,(ec,es):
        print "%s: IOError: %s [%d], skipping file" % (fn,es,ec)
        break
      line1=f.readline()
      if mailhdr(line1):
        print fn+": mailbox detected"
        if SAVECLEAN:
          rv=fcntl.fcntl(f.fileno(), fcntl.F_SETFL, os.O_NDELAY)
          rv=fcntl.fcntl(f.fileno(), fcntl.F_SETLKW,
                         struct.pack('hhqqhh', fcntl.F_WRLCK, 0,0,0,0,0))
          fc=open(fn+SAVECLEAN,"w")
          os.chmod(fn+SAVECLEAN,0700)
        mailcount=0
        box_infected=0
        while 1:
          mailcount+=1
          frm=line1
          lines=StringIO()
          while 1:
            line1=f.readline()
            if line1=="": break
            if mailhdr(line1): break
            lines.write(line1)
          if not scan(lines.getvalue(),frm.strip(),mailcount):
            if SAVECLEAN:
              fc.write(frm)
              fc.write(lines.getvalue())
          else:
            infected+=1
            box_infected+=1
          total+=1
          lines.close()
          if line1=="": break
        # remove .clean file if no viruses found
        if SAVECLEAN:
          os.chdir(cwd)
          if box_infected>0:
            # mailbox cleaned, viruses found
            fc.close()
            os.lseek(f.fileno(),0,0)
            os.ftruncate(f.fileno(),0)
            f.write(open(fn+SAVECLEAN,'r').read())
          try:
            os.unlink(fn+SAVECLEAN)
          except OSError:
            pass
      elif re.search('^NOOP CONNECT FROM: ',line1):
        print fn+": sagator's quarantine file detected"
        # skip to "data"
        while 1:
          line1=f.readline()
          if re.compile("^DATA",re.IGNORECASE).search(line1):
            break
        lines=StringIO()
        while 1:
          line1=f.readline()
          if (line1=='.\n') | (line1=='.\r\n'):
            break
          lines.write(line1)
        if scan(lines.getvalue(),os.path.basename(fn)):
          infected+=1
        total+=1
        lines.close()
        break # skip to next file
      else:
        # not mailbox
        debug.echo(2, "Scanning file: "+fn)
        if scan(open(fn,'r').read(),fn):
          infected+=1
        total+=1
      f.close()
      break
    except IOError,(ec,es):
      if ec!=35: # Resource deadlock avoided
        raise
      debug.echo(0, "Deadlock avoided, repeating last scan...")
print "Total infected/spamed files/emails:",\
  str(infected)+"/"+str(total),"in:", \
  time.mktime(time.localtime())-time0,"seconds"



'''
Basic realscanners for sagator, version 0.6.1

(c) 2003-2004 Jan ONDREJ (SAL) <ondrejj(at)salstar.sk>

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
    
'''

from avlib import *
import re,string,os,time,socket

__all__=['regexp_scan','string_scan','smtp_comm',
         'const','max_file_size','exec_any']

def iret(level,vname,ret):
    if vname:
      return level,vname,ret
    else:
      return 0.0,vname,ret

class regexp_scan(ascanner):
  '''
  Primitive regexp pattern scanner.
  
  Usage: regexp_scan({'VirName':['RegExp_Pattern...',...],...} [,size])

  Where: 'VirName' is a string, which identifies defined virus
         'RegExp_Pattern...' is a regexp pattern
         size is a number, which defines, how many bytes may be checked.
           If it is 0 or not defined, whole stream is scanned.
           If it is -1, email header is scanned.
    
  There can be more patterns for one virus. All patterns
  must match to assign an stream as virus.
  There can also be more virnames in one dictionary.
  '''
  name='RegExpScanner()'
  def __init__(self,regexps,size=0):
      self.regexps=regexps
      self.size=size
  def scanstream(self,stream):
      for vname,regs in self.regexps.items():
        for reg in regs:
          if self.size>0:
            reg1=re.search(reg,stream[:self.size])
          elif self.size<0:
            reg1=re.search(reg,stream[:len(mail.header)])
          else:
            reg1=re.search(reg,stream)
          if reg1:
            if reg==regs[-1]:
              mail.addheader("X-Sagator-RegExp",vname)
              return iret(1.0,vname,['regexp: '+vname+' FOUND, '+reg+'\n'])
            else:
              break
      return 0.0,'',[]

class string_scan(ascanner):
  '''
  Primitive string pattern scanner.
  
  Usage: string_scan({'VirName':['Pattern...',...],...} [,size])

  Where: 'VirName' is a string, which identifies defined virus
         'Pattern...' is a string pattern
         size is a number, which defines, how many bytes may be checked.
           If it is 0 or not defined, whole stream is scanned.
           If it is -1, email header is scanned.
    
  There can be more patterns for one virus. All patterns
  must match to assign an stream as virus.
  There can also be more virnames in one dictionary.
  '''
  name='StringScanner()'
  def __init__(self,strings,size=0):
      self.strings=strings
      self.size=size
  def scanstream(self,stream):
      for vname,regs in self.strings.items():
        for reg in regs:
          if self.size>0:
            str1=string.find(stream[:self.size],reg)
          elif self.size<0:
            str1=string.find(stream[:len(mail.header)],reg)
          else:
            str1=string.find(stream,reg)
          if str1>=0:
            if reg==regs[-1]:
              mail.addheader("X-Sagator-StringScan",vname)
              return iret(1.0,vname,['string: '+vname+' FOUND, '+reg+'\n'])
          else:
            break
      return 0.0,'',[]

class smtp_comm(ascanner):
  '''
  Primitive regexp pattern scanner for SMTP communication.
  
  Usage: smtp_comm({'VirName':['RegExp_Pattern...',...],...})

  Where: 'VirName' is a string, which identifies defined virus
         'RegExp_Pattern...' is a regexp pattern
    
  For more information about this scanner see documentation for
  regexp_scan realscanner.
  
  This scanner also can be used to apply some scnanner to defined
  email adresses. For example:

      match_all(
        smtp_comm({'anydomain:':['^RCPT TO:.*anydomain.dom']}),
        stream2mbox(libclam())
      )
  
  In this example, libclam() scanner is used only if mail is
  adressed to an user on anydomain.dom.
  '''
  name='smtp_comm()'
  def __init__(self,regexps):
      self.regexps=regexps
  def scanstream(self,stream):
      if globals.scan_only:
        return 0.0,'',[]
      for vname,regs in self.regexps.items():
        for reg in regs:
          if re.search(reg,mail.comm,re.IGNORECASE|re.MULTILINE):
            if reg==regs[-1]:
              mail.addheader("X-Sagator-SMTP-comm",vname)
              return iret(1.0,vname,['smtp_comm: '+vname+' FOUND, '+reg+'\n'])
          else:
            break
      return 0.0,'',[]
  def scanfile(self,files,dir=''):
      return self.scanstream('')

#####################################################################
### primitive class

class const(ascanner):
  '''
  const() realscanner
  
  Usage: const(level,'VirName',['virreturnstringline1',...])
      or const()
  
  Where: level is an float which defines returned infection level
           If level is not defined, after scanning an error is raised.
         'VirName',[...] is a returned virus.
  
  This scanner has no real functionality. It return always
  a defined virus, or clean (if virus is not defined).
  This scanner has no error codes.
  '''
  name='const()'
  ignore_name=True
  def __init__(self,level=None,vir='',ret=[]):
      self.level=level
      self.vir=vir
      self.ret=ret
      self.filename=''
  def scanstream(self,stream):
      if self.level==None:
        raise ScannerError,'Forced error by const() scanner'
      if self.filename:
        debug.echo(6,self.name,": Scanning file: ",self.filename)
      return self.level,self.vir,self.ret
  def scanfile(self,files,dirname=''):
      if self.level==None:
        raise ScannerError,'Forced error by const() scanner'
      for fn in files:
        debug.echo(6,self.name,": Scanning file: ",fn)
      return self.level,self.vir,self.ret

#####################################################################
### max_file_size class

class max_file_size(ascanner):
  '''
  max_file_size() realscanner
  
  Usage: max_file_size(size[,'name'])
  
  Where: size is a number of bytes, which if exceeded, virus is returned
         'name' is a string, which identifies the virus name.
            If it is not present, 'FileSizeOverrun' is returned.
  '''
  name='max_file_size()'
  def __init__(self,size,vname='FileSizeOverrun'):
      self.size=size
      self.vname=vname
  def scanstream(self,stream):
      if len(stream)<=self.size:
        return 0.0,'',[]
      else:
        return 1.0,self.vname,[self.filename+": "+self.vname]
  def scanfile(self,files,dir=''):
      for fn in files:
        if os.lstat(safe.fn(fn))[6]>self.size:
          return 1.0,self.vname,[fn+': '+self.vname]
      return 0.0,'',[]

#####################################################################
### exec_any class

class exec_any(ascanner):
  '''
  exec_any() realscanner
  
  Usage: exec_any(['command','arg1',...],[statusA1,...],[statusB1,...])
  
  Where: 'command' is a string, which command to run
         'argX' are command line arguments, as last argument the
           filename is added
         statusA1,... are integers, which defines "clean" exit codes
         statusB1,... are integers, which defines "virus found" exit codes
  
  If command returns exit status 0, no virus is returned.
  If exit code is any from statusX, virus is returned.
  Otherwise an error is raised.
  '''
  name='exec_any()'
  def __init__(self,arg,clean_status=[0],infected_status=[1]):
      self.arg=arg
      self.clean_status=clean_status
      self.infected_status=infected_status
  def scanfile(self,files,dir=''):
      for fn in files:
        pf=popen(self.arg+[fn])
        pf.wait()
        pf.close()
        if pf.exitstatus in self.clean_status:
          return 0.0,'',[]
        else:
          if pf.exitstatus in self.infected_status:
            return 1.0,'UnknownVirus',['ExitCode: '+str(pf.exitstatus)]
          else:
            raise ScannerError,'UnkonwnExitStatus: '+str(pf.exitstatus)



'''
mimeparse module, version 0.4.3

(c) 2003-2004 Jan ONDREJ (SAL) <ondrejj(at)salstar.sk>
                                                                                
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
'''

from avlib import *
import os,re,filetype
import email,mimetypes,email.Errors,email.Parser
import email.Utils,binascii,uu,base64
import multifile,mimetools
from StringIO import StringIO

__all__=['mimeparse','mime_ext','file_type']

class mime_ext_int(ascanner):
      '''
      Internal scanner for mime_ext.
      
      This scanner checks defined extension of file.
      Regexp check with ignoring of case is used.
      '''
      def __init__(self,wrong_type):
          self.filename=''
          self.wrong_type=wrong_type
      def scanstream(self,data):
          if (self.wrong_type!='') and (type(self.filename)==type('string')):
            if re.search(self.wrong_type,self.filename,re.IGNORECASE):
               return 1.0,'MatchedExtension', [self.filename+' matched '+self.wrong_type+'\n']
          return 0.0,'',[]

def decode_part(self,filename=''):
    '''
    Decode a part of message. Return at least some bytes from this part,
    if error occures.
    '''
    if self.is_multipart():
      return None,'multipart'
    cte = self.get('content-transfer-encoding', '').lower()
    debug.echo(5,"mimeparse(): decoding: ",filename,", type: ",cte)
    payload=self.get_payload()
    if cte == 'quoted-printable':
      return email.Utils._qdecode(payload),''
    elif cte == 'base64':
      try:
        return binascii.a2b_base64(payload),''
      except binascii.Error:
        # try another method to decode base64 stream
        sfp = StringIO()
        try:
          base64.decode(StringIO(payload+'\n'), sfp)
          return sfp.getvalue(),''
        except:
          debug.echo(2,"ERROR: mimeparse(): part_decode: Incorrect padding")
          return sfp.getvalue(),'truncated'
    elif cte in ('x-uuencode', 'uuencode', 'uue', 'x-uue'):
      sfp = StringIO()
      try:
        uu.decode(StringIO(payload+'\n'), sfp)
        return sfp.getvalue(),''
      except uu.Error:
        debug.echo(2,"ERROR: mimeparse(): part_decode: Some decoding problem")
        return sfp.getvalue(),'truncated'
    return payload,'encoded'

def decode_email_alt(data,scanner):
    '''
    Alternate method to decode an email.
    '''
    try:
      stream=StringIO(data)
      msg = mimetools.Message(stream)
      msgtype = msg.gettype()
      params = msg.getplist()
    except:
      debug.echo(1, "MimeParse: ERROR: mimetools.Message(), all decoders failed")
      debug.traceback(6, "MimeParse: ")
      return '',[],[]
    files=[]
    counter=0
    level,vir,ret=0.0,'',[]
    if msgtype[:10] == "multipart/":
      file = multifile.MultiFile(stream)
      file.push(msg.getparam("boundary"))
      while file.next():
        counter+=1
        submsg = mimetools.Message(file)
        sfn=submsg.getparam('name')
        if sfn:
          scanner.filename=normalize_filename(sfn)
        else:
          try:
            ext = mimetypes.guess_extension(submsg.gettype())
            if not ext:
              ext=".bin"
          except:
            ext=".bin"
          scanner.filename = 'part-%03d%s' % (counter,ext)
        debug.echo(5,"mimeparse(): decoding: ",scanner.filename)
        try:
          data = StringIO()
          mimetools.decode(file, data, submsg.getencoding())
        except binascii.Error:
          pass
        stream=data.getvalue()
        if stream:
          debug.echo(6,'mimeparse(): decoded len=',len(stream))
          if scanner.is_multistream:
            scanner.addstream(stream)
          else:
            l,v,r=scanner.scanstream(stream)
            if not vir:
              vir=v
            ret+=r
            level+=l
          files.append(scanner.filename)
      file.pop()
    if scanner.is_multistream:
      level,vir,ret=scanner.scanmultistream([])
    del scanner
    return level,vir,ret,files

def decode_email(data,scanner):
    '''
    Decode whole email and run a scanner on decoded parts.
    '''
    try:
      msg=email.message_from_string(data)
    except email.Errors.BoundaryError,es:
      debug.echo(2, "MimeParse: ERROR: BoundaryError: ",es)
      return decode_email_alt(data,scanner)
    except email.Errors.HeaderParseError,es:
      debug.echo(2, "MimeParse: ERROR: HeaderParseError: ",es)
      return decode_email_alt(data,scanner)
    files=[]
    counter=1
    level,vir,ret=0.0,'',[]
    for part in msg.walk():
      counter+=1
      try:
        sfn=part.get_filename()
        if sfn:
          scanner.filename=normalize_filename(sfn)
        else:
          # no filename, make own
          try:
            ext = mimetypes.guess_extension(part.get_type())
          except:
            ext='.bin'
          if not ext:
            ext='.bin'
          scanner.filename = 'part-%03d%s' % (counter, ext)
        stream,dstat=decode_part(part,scanner.filename)
        if stream!=None:
          debug.echo(5,"mimeparse(): stream len=",len(stream)," ",dstat)
      except:
        stream=""
      if stream:
        if scanner.is_multistream:
          scanner.addstream(stream)
        else:
          l,v,r=scanner.scanstream(stream)
          if not vir:
            vir=v
          ret+=r
          level+=l
        files.append(scanner.filename)
    if scanner.is_multistream:
      level,vir,ret=scanner.scanmultistream([])
    del scanner
    return level,vir,ret,files

class mimeparse(interscanner):
      '''
      mimeparse interscanner
      
      Usage: mimeparse(another_scanner(...))
      
      This scanner parses MIME attached mail and send multiple of
      streams into another_scanner(). If an multiscanner is used
      as another_scanner, optimal method is used (scans all files
      at once).
      '''
      name='MimeParse()'
      def __init__(self,scanner):
          self.name="MimeParse("+scanner.name+")"
          self.scanner=scanner
      def scanstream(self,stream):
          self.scanner.prescan()
          level,vir,ret,filelist=decode_email(stream,self.scanner)
          self.scanner.postscan(level,vir,ret)
          return level,vir,ret

class mime_ext(ascanner):
      '''
      mime_ext() is a extension scanner.

      Usage: mime_ext('wrong types regular expressios')
      
      This scanner checks defined extension of file in mime attachments.
      Regexp checks with ignoring of case are used.
      You can use predefined extensions from mime_ext.EXECUTABLES,
      for example:
        mime_ext(mime_ext.EXECUTABLES)
      '''
      name='MimeExt()'
      EXECUTABLES='(\.exe|\.com|\.pif|\.lnk|\.scr|\.vbs)$'
      def __init__(self,wrong_type=''):
          self.wrong_type=wrong_type
      def scanstream(self,stream):
          level,vir,ret,filelist=decode_email(stream,mime_ext_int(self.wrong_type))
          if vir:
            globals.found_by=self
          return level,vir,ret

class file_type(ascanner):
      '''
      file_type() is a realscanner, which scans type of a file
      
      Usage: file_type({'type':'vir',...})
      
      Where 'type' is a name of type returned by filetype.* function
            'vir' is a vir name returned, if file type is matched
      '''
      name='file_type()'
      def __init__(self,types):
          self.types=types
      def scanstream(self,stream):
          fta=filetype.what(stream)
          for ft in fta:
            if ft in self.types:
              return 1.0,self.types[ft],[self.filename+": "+str(fta)]
          return 0.0,'',[]
      def scanfile(self,files,dirname=''):
          for fname in files:
            fta=filetype.file(fname)
            for ft in fta:
              if ft in self.types:
                return 1.0,self.types[ft],[fname+": "+str(fta)]
          return 0.0,'',[]



'''
bogofilter module, version 0.3

(c) 2003 Jan ONDREJ (SAL) <ondrejj(at)salstar.sk>
                                                                                
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

'''

from avlib import *
import re

__all__=['bogofilter']

class bogofilter(ascanner):
  '''
  BogoFilter scanner
  
  Usage: bogofilter(['/usr/bin/bogofilter','-v'],SPAMSTRING,MAX_FILE_SIZE)
  
  Where: in [] is a path and argumets of bogofilter binary
         SPAMSTRING is a string, which define a report string,
           which is checked. If it is found, mail is assigned as 'SPAM'
         MAX_FILE_SIZE is a number, which defines maximum file size,
           which can be tested. If a mail (with header) is larger,
           it is not checked (by default 500000).

  This scanner scans for spams. If a spam is returned, a string
  SPAM is returned as virus name.
  '''
  is_spamscan=1
  name='BogoFilter()'
  def __init__(self,arg,spamstring='',max_file_size=500000):
      self.arg=arg
      self.SPAMSTRING=spamstring
      self.MAX_FILE_SIZE=max_file_size
  def scanstream(self,stream):
      exitstatus=''
      if len(stream)>self.MAX_FILE_SIZE:
        return 0.0,'',['File too long to test it. Size: '+str(len(stream))]
      pf=popen(self.arg)
      pf.write(stream)
      pf.tocmd.close()
      pf.wait()
      ret=pf.readlines()[0].rstrip()
      mail.addheader(ret)
      reg1=re.search("^X-Bogosity: ([A-Za-z]*), tests=([^,]*), spamicity=([-0-9.]*), version=(.*)$",ret)
      level,detected=0.0,''
      if reg1:
        level=float(reg1.group(3))
        if reg1.group(1)=="Yes":
          detected='SPAM'
      debug.echo(4,"bogofilter: ",ret)
      return level,detected,[ret]



'''
webscan.py, sagator's web scanner

(c) 2003-2004 Jan ONDREJ (SAL) <ondrejj(at)salstar.sk>
 
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

Add these lines to apache configuration:
  AddHandler mod_python .py
  PythonHandler mod_python.publisher
  PythonPath "sys.path+['/usr/lib/sagator']"

If you need another configuration for sagator, please copy your
sagator to another directory and change it's config file.

'''

from aglib import *
import types

# load config
from default import *
try:
  from config import *
except ImportError,es:
  if str(es)[-6:]=="config":
    print "ERROR! Config file not found! Exiting now."
    sys.exit(1)
  else:
    raise

def form(req,num=1):
    req.write('<form action="webscan.py" method=post enctype="multipart/form-data">\n')
    for i in range(num,0,-1):
      req.write('<input type=file name=file%04d size=80><br>\n' % i)
    req.write('<input type=submit>\n</form>')
    return ''

class io:
  def __init__(self,req):
      self.req=req
  def write(self,s):
      self.req.write(s+"<br>")

def index(req,**args):
    req.content_type='text/html'
    req.write("<h3>WebSagator %s-%s</h3>\n" % (version.VERSION,version.RELEASE))
    if req.method == 'GET':
      try:
        return form(req,int(args['n']))
      except:
        return form(req)
    debug.set_level(1)
    debug.logfd=io(req)
    for scnr in SCANNERS:
      scnr.reinit()
    for file in args.values():
      if type(file) == types.InstanceType:
        globals.reset()
        mail.filename=file.filename
        mail.data=file.file.read()
        mail.comm="NOOP WEBSAGATOR, remote host: %s" \
          % (req.get_remote_host(2))
        d=''
        for scnr in SCANNERS:
          level,detected,virlist,scan_reply,err=do_scan(scnr)
          if is_infected(level,detected):
            d="[%s,%0.2f]" % (globals.found_by.name,level)
            break
        if d:
          req.write("%s: <font color=red>%s</font> %s<br>\n" \
            % (mail.filename,detected,d))
        else:
          req.write("%s: no virus detected<br>\n" % mail.filename)
    return ''



'''
smtpd.py - SMTP daemon service for sagator.

(c) 2003-2004 Jan ONDREJ (SAL) <ondrejj(at)salstar.sk>

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

'''

from aglib import *
import signal,time
import stats

__all__=['smtpd']

class smtps(smtp):
  '''
  SMTP server class
  '''
  reg_smtp_reply=re.compile("^[0-9]{3} ")
  reg_smtp_reply_ok=re.compile("^[23][0-9][0-9] ")
  reg_xforward=re.compile("^XFORWARD ",re.IGNORECASE)
  reg_mailfrom=re.compile("^MAIL +FROM:(.*)$",re.IGNORECASE)
  reg_rcptto=re.compile("^RCPT +TO:(.*)$",re.IGNORECASE)
  reg_rset=re.compile("^RSET",re.IGNORECASE)
  reg_data=re.compile("^DATA",re.IGNORECASE)
  reg_quit=re.compile("^QUIT",re.IGNORECASE)
  def __init__(self,scanners,conn,addr):
    self.SCANNERS=scanners
    self.f=conn.makefile('rw',self.bufsize)
    self.write=conn.sendall
    self.readline=self.f.readline
    self.RECV=RECV_SMTP
    self.t0=time.time()
    self.srv=smtpc()
    self.write(self.srv.readline())
    mail.__init__()
    mail.addr='NOOP CONNECT FROM: '+addr[0]+':'+str(addr[1])\
      +', '+mail.getnamebyaddr(addr[0])+'\r\n'
  def send_ok(self,r="Ok"):
    self.write("250 "+r+"\r\n")
  def quit_connect(self):
    debug.echo(5,"SMTPS: Closing connect to server")
    self.RECV=RECV_QUIT
    self.srv.close()
    return self.recv()
  def check_reply(self,line):
    '''Check for SMTP reply. Return True, if 2XX or 3XX returned.'''
    self.srv.conn.sendall(line)
    # send reply
    while 1:
      l=self.srv.readline()
      if not l:
        debug.echo(3,"SMTPS: Connection closed after:",line.rstrip())
        return False
      debug.echo(7,'SMTPC: '+l)
      self.write(l)
      if self.reg_smtp_reply.search(l):
        break
    if self.reg_smtp_reply_ok.search(l):
      return True
    debug.echo(4,"SMTPS: "+l.rstrip()+' --- '+line.rstrip())
    return False
  def recv(self):
    try:
      line=self.readline()
    except IOError,(ec,es):
      debug.echo(0, "smtps(): ",es)
      return 1
    if not line: return 1
    if self.RECV: # receiving data or header
      if self.RECV==RECV_QUIT: # waiting for quit?
        if self.reg_quit.search(line):
          self.write("221 Bye\r\n")
          self.f.close()
          return 1
        else:
          self.write("502 SMTP communication stopped by Sagator\r\n")
          return 0
      elif (self.RECV==RECV_HEADER) & ((line=='\r\n') | (line=='\n')):
        # receiving header
        debug.echo(5,"SMTPS: HEADER DONE")
        mail.df.write(line)
        self.RECV=RECV_BODY
      elif (line=='.\r\n') | (line=='.\n'):
        # end of DATA part
        mail.close()
        if len(mail.data)==len(mail.header):
          debug.echo(0,"SMTPS: Warning: No header?")
          mail.header=''
        debug.echo(3,"SMTPS: BODY DONE, size: ",len(mail.data)," B")
        self.RECV=RECV_SMTP
        v,virname=checkvir(self.SCANNERS)
        self.tw=time.time()-self.t0
        if v!=S_TEMPFAIL:
          debug.echo(2,"STATS: "+str(self.tw)+
            " seconds, ",len(mail.data)," bytes, status: ",virname)
        if (v==S_OK) | (v==S_FORCE_SEND):
          debug.echo(5,"SMTPS: Sending data")
          self.srv.conn.sendall(mail.header)
          self.srv.conn.sendall(mail.xheader)
          self.srv.conn.sendall(mail.data[mail.bodypos:])
          debug.echo(1,"SMTPS: OK: 250 Ok")
          self.srv.conn.sendall(line) # send .
          self.write(self.srv.readline())
          stats.update(len(mail.data),self.tw)
          mail.__init__() # reinit
        elif v==S_REJECT:
          debug.echo(1,"SMTPS: REJECT: ","550 Content rejected - VIRUS ",virname)
          self.write("550 Content rejected - VIRUS "+str(virname)+"\r\n")
          stats.update(len(mail.data),self.tw,1)
          return self.quit_connect()
        elif v==S_DROP:
          debug.echo(1,"SMTPS: DROP: 250 mail dropped, VIRUS ",virname)
          self.write("250 mail dropped, VIRUS "+str(virname)+"\r\n")
          stats.update(len(mail.data),self.tw,1)
          return self.quit_connect()
        else: # S_TEMPFAIL
          debug.echo(1,"SMTPS: TEMPFAIL: 451 ",virname)
          self.write("451 "+str(virname)+"\r\n")
          stats.update(0,0,0,1) # update fail statistics
          return self.quit_connect()
      else:
        if self.RECV==RECV_HEADER:
          mail.hf.write(line)
        else:
          mail.df.write(line)
    else:
      mail.comm=mail.comm+line
      rcptto=self.reg_rcptto.search(line)
      mailfrom=self.reg_mailfrom.search(line)
      if rcptto:
        if self.check_reply(line):
          debug.echo(2,'SMTPS: '+line.rstrip())
          try:
            mail.recip.append(rfc822.parseaddr(rcptto.group(1))[1])
          except:
            pass
        return 0
      elif mailfrom:
        if self.check_reply(line):
          debug.echo(2,'SMTPS: '+line.rstrip())
          mail.sender=mailfrom.group(1)
      elif self.reg_rset.search(line):
        if self.check_reply(line):
          debug.echo(2,'SMTPS: '+line.rstrip(),", Resetting mail class")
          mail.__init__() # reinit
      elif self.reg_data.search(line):
        if self.check_reply(line):
          debug.echo(5,'SMTPS: '+line.rstrip())
          self.RECV=RECV_HEADER
      elif self.reg_quit.search(line):
        debug.echo(5,'SMTPS: '+line.rstrip())
        self.check_reply(line) # 221 Bye
        return 1
      elif self.reg_xforward.search(line):
        debug.echo(2,'SMTPS: ',line.rstrip())
        self.check_reply(line)
      else:
        debug.echo(6,'SMTPS: '+line.rstrip())
        self.check_reply(line)
    return 0

class smtpd(service):
  '''
  smtpd() service for sagator.
  
  This service can be used to start sagator as separate filtering SMTP
  daemon. Is is useful for postfix and any other SMTP daemon, which
  can use these filters.
  
  Usage: smtpd(scanners, host, port [,prefork])
  
  Where: scanners is an array of scanners (see README.scanners for more info)
         host is a an ip address to bind
         port is a port to bind
         prefork is a number, which defines preforked process count
  
  Example: smtpd('127.0.0.1', 27)
  '''
  name='smtpd()'
  def __init__(self,scanners,host,port,prefork=5,listen=5):
      self.SCANNERS=scanners
      self.HOST=host
      self.PORT=port
      self.LISTEN=listen
      self.PREFORK=prefork
      self.EXITING=False
  def main(self):
      test_scanners(self.SCANNERS)
      # bind to socket
      self.s=socket.socket(socket.AF_INET,socket.SOCK_STREAM)
      self.s.setsockopt(socket.SOL_SOCKET,socket.SO_REUSEADDR,1)
      try:
        self.s.bind((self.HOST,self.PORT))
        self.s.listen(self.LISTEN)
      except socket.gaierror,(ec,es):
        debug.echo(0,"smtpd(): ERROR: BIND ",es)
        sys.exit(1)
      # and now start the daemon
      signal.signal(signal.SIGCHLD,self.sigchld)
      signal.signal(signal.SIGHUP,sighup)
      signal.signal(signal.SIGTERM,signal.SIG_DFL)
      pid=self.daemonize()
      if pid>0:
        self.s.close()
        signal.signal(signal.SIGCHLD,signal.SIG_DFL)
        return [pid]
      # chroot
      dochroot()
      debug.echo(1,"smtpd(): service started, waiting for connections ...")
      # prefork
      signal.signal(signal.SIGTERM,self.sigterm)
      signal.signal(signal.SIGHUP,self.sighup)
      self.PID=os.getpid()
      self.childs=[]
      for i in range(self.PREFORK):
        self.fork()
      debug.echo(3,"smtpd(): pids: ",str(self.childs))
      try:
        while True:
          time.sleep(60)
      except KeyboardInterrupt:
        self.sigterm(None,None) # kill childs
        raise
  def fork(self):
      try:
        pid=os.fork()
        if pid>0:
          self.childs.append(pid)
          return pid
      except OSError,(ec,es):
        debug.echo(1,"ERROR: Can't fork:",es)
        return -1
      signal.signal(signal.SIGCHLD,signal.SIG_DFL)
      signal.signal(signal.SIGHUP,sighup) # reopen log
      signal.signal(signal.SIGTERM,signal.SIG_DFL)
      time1,timec='',1
      while True:
        try:
          # accept
          conn,addr = self.s.accept()
          socket_settimeout(conn,120)
          # reinit scanners
          for scnr in self.SCANNERS:
            scnr.reinit()
          # generate ID
          time2=time.strftime("%Y%m%d-%H%M%S",time.localtime(time.time()))
          if time2!=time1:
            time1,timec=time2,1
          globals.id="%s-%04d-%05d-%s" % (time2,timec,os.getpid(),randomchars(6))
          debug.echo(1,"Connection from: "+addr[0]+" at "+\
            time.strftime("%c",time.localtime()))
          debug.echo(1,"Process id: %s" % globals.id)
          m=smtps(self.SCANNERS,conn,addr)
          while 1:
            if m.recv(): break
          conn.shutdown(2)
          debug.echo(1,"Closing connection.")
          conn.close()
        except SmtpcError,(ses,es):
          # error is written, accept new connection
          pass
        except socket.error,err:
          try:
            ec=err[0]
            es=err[1]
          except IndexError:
            ec=0
            es=str(err)
          if ec==107: # Transport endpoint is not connected
            debug.echo(1,"smtpd(): Connection closed by peer.")
          elif ec==104: # Connection reset by peer
            debug.echo(1,"smtpd(): Connection reset by peer.")
          elif ec==32: # 'Broken pipe'
            debug.echo(0,"smtpd(): ERROR: Broken pipe.")
          elif ec!=4:
            debug.echo(0,'smtpd(): ERROR: SocketError: ',es)
            debug.traceback(4,"smtpd(): ")
        except KeyboardInterrupt:
          self.sigterm(None,None)
        except:
          debug.echo(1,"ERROR: smtpd(): Unknow error!")
          debug.traceback(1,"smtpd(): ")
  def sighup(self,sn,stack):
      # reopen log for master process
      sighup(sn,stack)
      # reopen logs for all childs
      for pid in self.childs:
        try:
          os.kill(pid,signal.SIGHUP)
        except:
          pass
  def sigchld(self,sn,stack):
      # try to remove older zombies (sigchld lost?)
      try:
        while 1:
          pid,s=os.waitpid(0,os.WNOHANG)
          if pid==0:
            break
          if pid in self.childs:
            self.childs.remove(pid)
          debug.echo(1,"smtpd(): Child exited ... pid,status: ",(pid,s))
      except OSError: # no child processes?
        pass
      if not self.EXITING:
        # start another child
        debug.echo(1,"WARNING: smtpd(): An child exited? Starting new one ...")
        self.fork()
  def sigterm(self,sn,stack):
      self.EXITING=True
      if os.getpid()!=self.PID:
        debug.echo(1,"smtpd(child): Exiting - SIGTERM ...")
        os._exit(0)
      # kill all childs
      for pid in self.childs:
        debug.echo(5,'smtpd(master): killing children %d (TERM signal)' % pid)
        try:
          os.kill(pid,signal.SIGTERM)
        except OSError:
          pass
      time.sleep(0.1)
      if self.childs:
        time.sleep(1)
        for pid in self.childs:
          debug.echo(3,'smtpd(master): killing children %d (KILL signal)' % pid)
          try:
            os.kill(pid,signal.SIGKILL)
          except OSError:
            pass
      os._exit(0)



'''
filetype module, version 0.3.1

(c) 2003 Jan ONDREJ (SAL) <ondrejj(at)salstar.sk>
                                                                                
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

String constants are based on Magdir/* from file package.
Extended to better support sfx archives.
 
Function returns an array of filetypes. For example for zip/sfx is returned:
  ['zip','exe'] in this order.
'''

import re,avlib

class rsc:
      def __init__(self,ext,*res):
          self.ext=ext
          self.fx=[]
          for r in res:
            self.fx.append(re.compile(r,re.DOTALL))
      def check(self,string):
          for i in self.fx:
            if i.search(string):
              return 1
          return 0

class rsc_multipos(rsc):
      def __init__(self,ext,res,*pos):
          self.ext=ext
          self.fx=[]
          for n in pos:
            self.fx.append(re.compile(
              re.sub('\{\}','{'+str(n)+'}',res),
              re.DOTALL))

type_table = [
  rsc('Z','^\037\235'),
  rsc('gz','^\037\213'),
  rsc('bz2','^BZh'),
  rsc('bz','^BZ'),
  rsc('lzop','^\x89\x4c\x5a\x4f\x00\x0d\x0a\x1a\x0a'),
  rsc('tar','^.{257}ustar(\0|\040\040\0)'),
  rsc('zip','^PK\003\004'),
  rsc('zoo','^.{20}\xdc\xa7\xc4\xfd'),
  rsc('rar','^Rar!'),
  rsc('arj','^\x60\xea'),
  rsc('lha','^..-l[hz].-'),
  rsc('uc2','^UC2\x1a'),
  rsc('cab','^MSCF\0\0\0\0'),
  rsc('deb','^!<arch>\ndebian'),
  rsc('cpio','^(07070[712]|\xc7\x71|\x71\xc7)'),
  rsc('arc','^\x1a[\x01\x02\x03\x04\x05\x06\x07\x08\x09]'),
  rsc('rpm','^\xab\xed\xdb\xee'),
  rsc('uu','^begin\040'),
  # Executables
  rsc_multipos('zip','^MZ.{}PK\003\004',
               11694,13297,15588,15770,28374,25115,26331,47031,49845,69120),
  rsc_multipos('rar','^MZ.{}Rar!',7193,38910,51710,95110),
  rsc('arj','^MZ.{28}RJSX\xff\xff'),
  rsc('lha','^MZ.{34}LHa\'s SFX','^MZ.{1636}-lh5-'),
  rsc('winpe''^.{128}PE\0\0'),
  rsc('exe','^[LM]Z'),
  rsc('elf','^\177ELF')
]

def what(stream):
    r=[]
    for f in type_table:
      if f.check(stream):
        r.append(f.ext)
    return r

def file(filename):
    return what(avlib.safe.open(filename,'rb').read(99999))



#!/usr/bin/python2

'''
collector.py - statistics collector service for sagator

(c) 2003-2004 Jan ONDREJ (SAL) <ondrejj(at)salstar.sk>

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

'''

from aglib import *
from stats import *

__all__=['collector']

class collector(service):
  '''
  Statistics collector service for sagator.
  
  This service can be used to collect some data for statistics.
  
  Usage: collector(name,port)
  
  Example: collector('0.0.0.0',28)
  '''
  name='collector()'
  def __init__(self,host,port):
      self.HOST=host
      self.PORT=port
  def main(self):
      globals.stat_coll=stats.collector((self.HOST,self.PORT))
      debug.dup()
      pid=globals.stat_coll.main()
      if pid>0:
        return [pid]
      else:
        return []
